############################################################
# Maximum Likelihood for Ordered model with Latent Class
# By: Mauricio Sarrias
############################################################

# Helper function for time
my.time <- function(object){
  et <- object[3]
  s <- round(et,0)
  h <- s%/%3600
  s <- s-3600*h
  m <- s%/%60
  s <- s-60*m
  paste(h, "h:", m, "m:", s, "s", sep = "")
}

jacobian <- function(kappa){
  ## ij elemenet is dkappa_i/d alpha_j  
  k <- length(kappa)
  etheta <- exp(kappa)
  mat <- matrix(0 ,k, k)
  for (i in 1:k) mat[i:k, i] <- etheta[i]
  mat
}


## suml function from mlogit (Croissant, 2013)
suml <- function(x){
  n <- length(x)
  if (!is.null(dim(x[[1]]))){
    d <- dim(x[[1]])
    s <- matrix(0,d[1], d[2])
    for (i in 1:n){
      x[[i]][is.na(x[[i]])] <- 0
      s <- s + x[[i]]
    }
  }
  else{
    s <- rep(0,length(x[[n]]))
    for (i in 1:n){
      x[[i]][is.na(x[[i]])] <- 0
      s <- s + x[[i]]
    }
  }
  s
}

repRows <- function(x, n){
  matrix(rep(x, each = n), nrow = n)
}

repCols <- function(x, n){
  matrix(rep(x, each = n), ncol = n, byrow = TRUE)
}

## ml function
ml_ord_lc <- function(theta, y, X, H, Q, id = NULL, weights = NULL,
                      gradient = TRUE, get.bi = TRUE){
  # theta: 
  # -kappas: (J-2) * Q
  # -betas: Q * K

  
  # Parameters and globals
  K <- ncol(X)
  N <- nrow(X)
  panel <- !is.null(id)
  if (panel) if (length(weights == 1)) weights <- rep(weights, N)
  
  # For gradient
  ## it is assumed that y is a "factor" class
  y <- as.factor(y)
  m       <- sort(unique(y))
  m       <- unclass(m)
  J       <- length(levels(y))
  ## Just J-2 alpha are estimated, so the gradient is just for 2:(J-1) cuts
  m       <- as.matrix(m[2:(J - 1)], nrow = (J - 2))               
  y       <- unclass(y)
  delta   <- t(kronecker(m, t(rep(1, nrow(X)))))    
  ## indicator for alternative and previous alternative
  deltaj  <- delta == y
  deltak  <- delta == y - 1
  
  # Parameters
  alpha    <- matrix(theta[1L:((J - 2) * Q)], nrow = (J - 2), ncol = Q)
  kappa    <- c()
  for (q in 1:Q) kappa <- cbind(kappa, c(-Inf, cumsum(c(0, exp(alpha[, q]))) , +Inf))
  
  beta <- matrix(theta[((J - 2) * Q + 1):((J - 2) * Q + K * Q)], nrow = K, ncol = Q)
  rownames(beta) <- colnames(X); colnames(beta) <- paste("class", 1:Q, sep = ":")
  gamma <- theta[-c(1L:((J - 2) * Q + K * Q))]
  
  # Make weights from a multinomial logit model
  ew <- lapply(H, function(x) exp(crossprod(t(x), gamma)))
  sew <- suml(ew)
  Wnq <- lapply(ew, function(x){ v <- x / sew;
                                v[is.na(v)] <- 0;
                                as.vector(v)})
  Wnq <- Reduce(cbind, Wnq) # W
  
  # Make Probit Probability
  pfun <- pnorm
  dfun <- dnorm
  
  index <- tcrossprod(X, t(beta))
  ck1 <- c()
  ck2 <- c()
  for (q in 1:Q) {
    ck1 <- cbind(ck1, kappa[y + 1, q])
    ck2 <- cbind(ck2, kappa[y, q])
  }
  eta1     <- pmin(ck1 - index, 100)
  eta2     <- pmax(ck2 - index, -100)
  Pntq <- pfun(eta1) - pfun(eta2)
  if (panel) Pnq <- apply(Pntq, 2, tapply, id, prod) else Pnq <- Pntq
  WPnq <- Wnq * Pnq 
  Ln <- pmax(apply(WPnq, 1, sum), .Machine$double.eps)
  if (get.bi) Qnr <- WPnq / Ln
  lnL <- if (panel) sum(log(Ln) * weights[!duplicated(id)]) else sum(log(Ln) * weights)
  
  ## Gradient
  if (gradient) {
    phi1     <- pmax(dfun(eta1), .Machine$double.eps)
    phi2     <- pmax(dfun(eta2), .Machine$double.eps)
    lambda   <- (phi2 - phi1) / pmax(Pntq, .Machine$double.eps) # N * Q
    Qnr <- WPnq / Ln
    if (panel) Qnr <- Qnr[id, ]
    eta <- lambda * Qnr
    etar <- eta[, rep(1:Q, each = K)]
    Xg <- X[, rep(1:K, Q)] # N * (K * Q)
    grad.beta <- Xg * etar
    
    ## Gradient for kappas
    etak    <- Qnr[, rep(1:Q, each = (J - 2))]
    deltaj  <- deltaj[, rep(1:(J - 2), Q)]
    deltak  <- deltak[, rep(1:(J - 2), Q)]
    pitqk   <- pmax(Pntq[, rep(1:Q, each = (J - 2))], .Machine$double.eps)
    phi1k   <- phi1[, rep(1:Q, each = (J - 2))]
    phi2k   <- phi2[, rep(1:Q, each = (J - 2))]
    lambdak <- (deltaj *  phi1k - deltak * phi2k) / pitqk
    gkappa  <- etak * lambdak
    # Create block jacobian matrix
    Bjac   <- matrix(0, nrow = (J - 2) * Q, ncol = (J - 2) * Q)
    for (q in 1:Q) {
      Bjac[((J - 2)*q - (J - 3)):(q*(J - 2)), 
           ((J - 2)*q - (J - 3)):(q*(J - 2))] <- jacobian(alpha[, q])
    }
    gkappa <- gkappa %*% Bjac
    
    # Grad for gamma
    if (panel) {
      Wnq <- Wnq[id, ]
      H <- lapply(H, function(x) x[id, ])
    }
    Wg <- vector(mode = "list", length = Q)
    IQ <- diag(Q)
    for (q in 1:Q) Wg[[q]] <- rowSums(Qnr * (repRows(IQ[q, ], N) - repCols(Wnq[, q], Q)))
    grad.gamma <- suml(mapply("*", H, Wg, SIMPLIFY = FALSE))
    gari <- cbind(gkappa, grad.beta, grad.gamma)
    colnames(gari) <- names(theta)
    attr(lnL, "gradient") <- gari * weights
  }
  if (get.bi) {
    attr(lnL, "bi")  <- beta
    attr(lnL, "Qir") <- Qnr
    attr(lnL, "Wnq") <- Wnq
  }
  lnL
}

has.intercept.default <- function(object, ...) {
  has.intercept(formula(object), ...)
}


has.intercept.formula <- function(object, ...) {
  attr(terms(object), "intercept") == 1L
}

has.intercept.Formula <- function(object, rhs = NULL, ...) {
  if (is.null(rhs)) rhs <- 1:length(attr(object, "rhs"))
  sapply(rhs, function(x) has.intercept(formula(object, lhs = 0, rhs = x)))
}


ord_lc <- function(formula, data, subset, na.action,
                   start = NULL, Q = 2, panel = FALSE, 
                   index = NULL, print.init = FALSE, 
                   gradient = TRUE, ...){
  require(Rchoice)
  require(plm)
  require(Formula)
  start.time <- proc.time()
  callT <- match.call(expand.dots = TRUE)
  callF <- match.call(expand.dots = FALSE)
  
  formula <- as.Formula(formula)
  nframe  <- length(sys.calls())
  
  if (is.null(callT$method)) callT$method <- 'bfgs'
  
  if (!is.null(index) && !panel) {
    panel <- TRUE
    warning("panel was FALSE, while index was not NULL. Panel is now TRUE")
  }
  if (panel) {
    if (inherits(data, "pdata.frame") && !is.null(index))
      warning("The index argument is ignored because data is a pdata.frame")
    if (!inherits(data, "pdata.frame") && is.null(index)) 
      stop("The data is not pdata.frame and index is needed")
    if (!inherits(data, "pdata.frame")) data <- pdata.frame(data, index)
  }
  
  ## Model frame
  mf <- callT
  m  <- match(c("formula", "data", "subset", "na.action", "weights"), names(mf), 0L)
  mf <- mf[c(1L, m)]
  mf$formula <- formula
  mf[[1L]] <- as.name("model.frame")
  mf$data <- data
  mf <- eval(mf, parent.frame())
  
  ## Index
  index <- attr(data, "index")
  if (!is.null(index)) rownames(index) <- rownames(mf)
  index <- index[rownames(mf), ]
  index <- data.frame(lapply(index , function(x) x[drop = TRUE]), row.names = rownames(index))
  
  ## check Panel
  if (panel) {
    id <- index$id
    if (is.null(id)) stop("No individual index")
  } else {
    id <- NULL
  }
  
  ## Extract elements of the model
  y <- model.response(mf)
  X <- model.matrix(formula, mf, rhs = 1) 
  init <- Rchoice(as.formula(formula(formula, rhs = 1)), 
                  data = mf, 
                  family = ordinal('probit'))
  
  ## Create matrix H
  indata <- if (is.null(id)) mf else mf[!duplicated(id), ]
  cldata <- indata[rep(seq_len(nrow(indata)), each = Q), ] #expand H
  if (is.null(id)) {
      class   <- factor(rep(1:Q, nrow(X)))
      N <- length(y)  
  } else {
      id.un <- unique(id)
      class   <- factor(rep(1:Q, length(id.un)))
      N <- length(id.un)
  }
  cldata   <- cbind(cldata, class, row.names = NULL)
  class.var <- formula(formula, rhs = 2, lhs = 0) 
  has.int <- has.intercept(class.var)
  if (has.int) intercept.char <- "factor(class)" else intercept.char <- NULL 
 if (!has.int) {
      class.var <- update(class.var, ~. + 1)
      class.var <- update(class.var, ~.)
      class.var.char <- as.character(class.var)[2]
      has.xclass <- as.character(class.var)[2]
      class.var.var <- colnames(model.matrix(update(class.var, ~. + 1), cldata))[-1]
      class.var.char <- paste("(", class.var.var, "):class", sep = "")
      if (class.var.char == "1") class.var.char <- class.var.var <- NULL
    } else {
      has.xclass <- as.character(class.var)[2]
      if (has.xclass == "1") {
        class.var.char <- NULL
      } else {
        class.var.var <- colnames(model.matrix(update(class.var, ~. + 1), cldata))[-1]
        class.var.char <- paste("(", class.var.var, "):class", sep = "")
      } 
    }
    resp.name <- as.character(attr(formula, "lhs"))
    form.char <- paste(c(intercept.char, class.var.char), collapse = "+")
    nformula <- as.formula(paste(resp.name, " ~ ", form.char))
    H <- model.matrix(nformula, cldata)[, -1, drop = F]
    
    lev1 <- levels(class)[1]
    lev1 <- paste("class", lev1, sep = "")
    if (has.xclass != "1") {
      toremove <- unlist(lapply(as.list(class.var.var), function(x) paste(lev1, x, sep = ":")))
      revtoremove <- unlist(lapply(as.list(class.var.var), function(x) paste(x, lev1, sep = ":")))
      toremove <- colnames(H) %in% c(toremove, revtoremove)
      H <- H[, !toremove, drop = FALSE]
    } 
    namesH <- colnames(H)
    for (i in 1:length(namesH)) namesH[i] <- sub('factor', '', namesH[i])
    colnames(H) <- namesH
    
    Hl <- vector(length = Q, mode = "list")
    names(Hl) <- levels(class)
    for (i in levels(class)) {
      Hl[[i]] <- H[class == i, , drop = FALSE]
    }
    
    ## Initial values
    alphas <- attr(coef(init), "alphas")
    mean   <- attr(coef(init), "fixed")
    init.shift <- seq(-0.02, 0.02, length.out = Q)
    lc.mean <- c()
    lc.alpha <- c()
    for (i in 1:Q) {
      lc.mean  <- c(lc.mean,  mean + init.shift[i])
      lc.alpha <- c(lc.alpha, alphas + init.shift[i])
    }
    lc.names <- c()
    lc.nalpha <- c()
    for (i in 1:Q) {
      lc.names <- c(lc.names, paste('class', i, names(mean), sep = '.'))
      lc.nalpha <- c(lc.nalpha, paste('class', i, names(alphas), sep = '.'))
    } 
    
    names(lc.mean) <- lc.names
    names(lc.alpha) <- lc.nalpha
    s_class <- rep(0, ncol(H))
    names(s_class) <- colnames(H)
    start <- c(lc.alpha, lc.mean, s_class)
    print(start)
    
    # call to maxLik with the relevant arguments
    ml <- callT
    m <- match(c("print.level", "ftol", "tol", "reltol",
                 "gradtol", "steptol", "lambdatol", "qrtol",
                 "iterlim", "fixed", "activePar", "method"), names(ml), 0)
    ml <- ml[c(1, m)]
    
    weights = 1
    get.bi  = FALSE
    args <- list(as.name('y'), as.name('X'), as.name('id'), 
                 as.name('Hl'), as.name('Q'), as.name('weights'),
                 as.name('gradient'), as.name('get.bi'))
    ml[c('y', 'X', 'id', 'H', 'Q', 'weights', 'gradient', 'get.bi')] <- args
    ml$logLik <- as.name('ml_ord_lc')
    ml$start <- start
    ml[[1]] <- as.name('maxLik')
    result <- eval(ml, sys.frame(which = nframe))
    
    bhat <- coef(result)
    out2 <- ml_ord_lc(bhat, y = y, X = X, 
                      H = Hl, Q = Q, id = id, weights = weights, 
                      gradient = FALSE, get.bi = TRUE)
    Qic <- attr(out2, 'Qir')
    Wic <- attr(out2, 'Wnq')
    colnames(Qic) <- colnames(Wic) <- paste("class", 1:Q, sep = ":")
    
    
    result[c('call', 'args', 'model', 'Q', 'Qic', 'Wic', 'N')] <- list(callT, args, data, Q,  Qic, Wic, N)
    result
}
